const version = "0.0.2";

function RuleCreator(urlCreator) {
  return function createNamedRule({
    name,
    meta,
    ...rule
  }) {
    return createRule({
      meta: {
        ...meta,
        docs: {
          ...meta.docs,
          url: urlCreator(name)
        }
      },
      ...rule
    });
  };
}
function createRule({
  create,
  defaultOptions,
  meta
}) {
  return {
    create: (context) => {
      const optionsWithDefault = context.options.map((options, index) => {
        return {
          ...defaultOptions[index] || {},
          ...options || {}
        };
      });
      return create(context, optionsWithDefault);
    },
    defaultOptions,
    meta
  };
}
const createEslintRule = RuleCreator(
  () => "https://github.com/antfu/eslint-plugin-command"
);

const toFunction = {
  name: "to-function",
  match: /^[\/:@]\s*(to-(?:fn|function)|2f|tf)$/,
  action(ctx) {
    const arrowFn = ctx.findNodeBelow("ArrowFunctionExpression");
    if (!arrowFn)
      return ctx.reportError("Unable to find arrow function to convert");
    let start = arrowFn;
    let id;
    const body = arrowFn.body;
    if (arrowFn.parent.type === "VariableDeclarator" && arrowFn.parent.id.type === "Identifier") {
      id = arrowFn.parent.id;
      if (arrowFn.parent.parent.type === "VariableDeclaration" && arrowFn.parent.parent.kind === "const" && arrowFn.parent.parent.declarations.length === 1)
        start = arrowFn.parent.parent;
    } else if (arrowFn.parent.type === "Property" && arrowFn.parent.key.type === "Identifier") {
      id = arrowFn.parent.key;
      start = arrowFn.parent.key;
    }
    ctx.removeComment();
    ctx.report({
      node: arrowFn,
      loc: {
        start: start.loc.start,
        end: body.loc.start
      },
      message: "Convert to function",
      fix(fixer) {
        const code = ctx.context.sourceCode.text;
        const textName = id ? code.slice(id.range[0], id.range[1]) : "";
        const textArgs = arrowFn.params.length ? code.slice(arrowFn.params[0].range[0], arrowFn.params[arrowFn.params.length - 1].range[1]) : "";
        const textBody = body.type === "BlockStatement" ? code.slice(body.range[0], body.range[1]) : `{
  return ${code.slice(body.range[0], body.range[1])}
}`;
        const textGeneric = arrowFn.typeParameters ? code.slice(arrowFn.typeParameters.range[0], arrowFn.typeParameters.range[1]) : "";
        const textTypeReturn = arrowFn.returnType ? code.slice(arrowFn.returnType.range[0], arrowFn.returnType.range[1]) : "";
        const textAsync = arrowFn.async ? "async" : "";
        const fnBody = [`${textGeneric}(${textArgs})${textTypeReturn}`, textBody].filter(Boolean).join(" ");
        let final = [textAsync, `function`, textName, fnBody].filter(Boolean).join(" ");
        if (arrowFn.parent.type === "Property")
          final = [textAsync, textName, fnBody].filter(Boolean).join(" ");
        return fixer.replaceTextRange([start.range[0], arrowFn.range[1]], final);
      }
    });
  }
};

const toArrow = {
  name: "to-arrow",
  match: /^[\/:@]\s*(to-arrow|2a|ta)$/,
  action(ctx) {
    const fn = ctx.findNodeBelow("FunctionDeclaration", "FunctionExpression");
    if (!fn)
      return ctx.reportError("Unable to find function declaration to convert");
    const id = fn.id;
    const body = fn.body;
    let rangeStart = fn.range[0];
    const rangeEnd = fn.range[1];
    ctx.removeComment();
    ctx.report({
      node: fn,
      loc: {
        start: fn.loc.start,
        end: body.loc.start
      },
      message: "Convert to arrow function",
      fix(fixer) {
        const code = ctx.context.sourceCode.text;
        let textName = id ? code.slice(id.range[0], id.range[1]) : "";
        const textArgs = fn.params.length ? code.slice(fn.params[0].range[0], fn.params[fn.params.length - 1].range[1]) : "";
        const textBody = body.type === "BlockStatement" ? code.slice(body.range[0], body.range[1]) : `{
  return ${code.slice(body.range[0], body.range[1])}
}`;
        const textGeneric = fn.typeParameters ? code.slice(fn.typeParameters.range[0], fn.typeParameters.range[1]) : "";
        const textTypeReturn = fn.returnType ? code.slice(fn.returnType.range[0], fn.returnType.range[1]) : "";
        const textAsync = fn.async ? "async" : "";
        let final = [textAsync, `${textGeneric}(${textArgs})${textTypeReturn} =>`, textBody].filter(Boolean).join(" ");
        if (fn.type === "FunctionDeclaration" && textName) {
          final = `const ${textName} = ${final}`;
        } else if (fn.parent.type === "Property") {
          rangeStart = fn.parent.range[0];
          textName = code.slice(fn.parent.key.range[0], fn.parent.key.range[1]);
          final = `${textName}: ${final}`;
        } else if (fn.parent.type === "MethodDefinition") {
          rangeStart = fn.parent.range[0];
          textName = code.slice(fn.parent.key.range[0], fn.parent.key.range[1]);
          final = `${textName} = ${final}`;
        }
        return fixer.replaceTextRange([rangeStart, rangeEnd], final);
      }
    });
  }
};

const keepSorted = {
  name: "keep-sorted",
  match: /^[\/@:]\s*(keep-sorted|sorted)$/,
  action(ctx) {
    const node = ctx.findNodeBelow(
      "ObjectExpression",
      "ArrayExpression",
      "TSInterfaceBody",
      "TSTypeLiteral"
    );
    if (!node)
      return ctx.reportError("Unable to find object/array/interface to sort");
    const firstToken = ctx.context.sourceCode.getFirstToken(node);
    const lastToken = ctx.context.sourceCode.getLastToken(node);
    if (!firstToken || !lastToken)
      return ctx.reportError("Unable to find object/array/interface to sort");
    function sort(list, getName) {
      if (list.length < 2)
        return;
      const reordered = list.slice();
      const ranges = /* @__PURE__ */ new Map();
      const names = /* @__PURE__ */ new Map();
      const rangeStart = Math.max(
        firstToken.range[1],
        ctx.context.sourceCode.getIndexFromLoc({
          line: list[0].loc.start.line,
          column: 0
        })
      );
      let rangeEnd = rangeStart;
      for (let i = 0; i < list.length; i++) {
        const item = list[i];
        const name = getName(item);
        names.set(item, name);
        let lastRange = item.range[1];
        const endStartToken = ctx.context.sourceCode.getTokenByRangeStart(lastRange);
        if (endStartToken?.type === "Punctuator" && endStartToken.value === ",")
          lastRange = endStartToken.range[1];
        if (ctx.context.sourceCode.getText()[lastRange] === "\n")
          lastRange++;
        ranges.set(item, [rangeEnd, lastRange]);
        rangeEnd = lastRange;
      }
      const segments = [];
      let segmentStart = -1;
      for (let i = 0; i < list.length; i++) {
        if (names.get(list[i]) == null) {
          if (segmentStart > -1)
            segments.push([segmentStart, i]);
          segmentStart = -1;
        } else {
          if (segmentStart === -1)
            segmentStart = i;
        }
      }
      if (segmentStart > -1 && segmentStart !== list.length - 1)
        segments.push([segmentStart, list.length]);
      for (const [start, end] of segments) {
        reordered.splice(
          start,
          end - start,
          ...reordered.slice(start, end).sort((a, b) => {
            const nameA = names.get(a);
            const nameB = names.get(b);
            return nameA < nameB ? -1 : nameA > nameB ? 1 : 0;
          })
        );
      }
      const changed = reordered.some((prop, i) => prop !== list[i]);
      if (!changed)
        return;
      const newContent = reordered.map((i) => {
        const range = ranges.get(i);
        return ctx.context.sourceCode.text.slice(range[0], range[1]);
      }).join("");
      ctx.report({
        node,
        loc: {
          start: node.loc.start,
          end: node.loc.end
        },
        message: "Keep sorted",
        fix(fixer) {
          return fixer.replaceTextRange([rangeStart, rangeEnd], newContent);
        }
      });
    }
    if (node.type === "ObjectExpression") {
      sort(
        node.properties,
        (prop) => {
          if (prop.type === "Property" && prop.key.type === "Identifier")
            return prop.key.name;
          return null;
        }
      );
    } else if (node.type === "ArrayExpression") {
      sort(
        node.elements.filter(Boolean),
        (element) => {
          if (element.type === "Identifier")
            return element.name;
          if (element.type === "Literal")
            return String(element.value);
          return null;
        }
      );
    } else if (node.type === "TSInterfaceBody") {
      sort(
        node.body,
        (prop) => {
          if (prop.type === "TSPropertySignature" && prop.key.type === "Identifier")
            return prop.key.name;
          return null;
        }
      );
    } else if (node.type === "TSTypeLiteral") {
      sort(
        node.members,
        (prop) => {
          if (prop.type === "TSPropertySignature" && prop.key.type === "Identifier")
            return prop.key.name;
          return null;
        }
      );
    }
  }
};

const FOR_TRAVERSE_IGNORE = [
  "FunctionDeclaration",
  "FunctionExpression",
  "ArrowFunctionExpression",
  "WhileStatement",
  "DoWhileStatement",
  "ForInStatement",
  "ForOfStatement",
  "ForStatement",
  "ArrowFunctionExpression"
];
const toForEach = {
  name: "to-for-each",
  match: /^[\/:@]\s*(?:to-|2)?(?:for-?each)$/i,
  action(ctx) {
    const node = ctx.findNodeBelow("ForInStatement", "ForOfStatement");
    if (!node)
      return ctx.reportError("Unable to find for statement to convert");
    const continueNodes = [];
    const result = ctx.traverse(node.body, (path, { STOP, SKIP }) => {
      if (FOR_TRAVERSE_IGNORE.includes(path.node.type))
        return SKIP;
      if (path.node.type === "ContinueStatement") {
        continueNodes.push(path.node);
      } else if (path.node.type === "BreakStatement") {
        ctx.reportError(
          "Unable to convert for statement with break statement",
          {
            node: path.node,
            message: "Break statement has no equivalent in forEach"
          }
        );
        return STOP;
      } else if (path.node.type === "ReturnStatement") {
        ctx.reportError(
          "Unable to convert for statement with return statement",
          {
            node: path.node,
            message: "Return statement has no equivalent in forEach"
          }
        );
        return STOP;
      }
    });
    if (!result)
      return;
    let textBody = ctx.source.getText(node.body);
    continueNodes.sort((a, b) => b.loc.start.line - a.loc.start.line).forEach((c) => {
      textBody = textBody.slice(0, c.range[0] - node.body.range[0]) + "return" + textBody.slice(c.range[1] - node.body.range[0]);
    });
    if (!textBody.trim().startsWith("{"))
      textBody = `{
${textBody}
}`;
    if (node.type === "ForOfStatement") {
      const left = node.left.type === "VariableDeclaration" ? node.left.declarations[0].id : node.left;
      const textLeft = ctx.source.getText(left);
      const textRight = ctx.source.getText(node.right);
      const str = `${textRight}.forEach(${textLeft} => ${textBody})`;
      ctx.removeComment();
      ctx.report({
        node,
        message: "Convert to forEach",
        fix(fixer) {
          return fixer.replaceText(node, str);
        }
      });
    } else if (node.type === "ForInStatement") {
      const textLeft = ctx.source.getText(node.left);
      const textRight = ctx.source.getText(node.right);
      const str = `Object.keys(${textRight}).forEach(${textLeft} => ${textBody})`;
      ctx.removeComment();
      ctx.report({
        node,
        message: "Convert to forEach",
        fix(fixer) {
          return fixer.replaceText(node, str);
        }
      });
    }
  }
};

const toForOf = {
  name: "to-for-of",
  match: /^[\/:@]\s*(?:to-|2)?(?:for-?of)$/i,
  action(ctx) {
    const target = ctx.findNodeBelow((node) => {
      if (node.type === "CallExpression" && node.callee.type === "MemberExpression" && node.callee.property.type === "Identifier" && node.callee.property.name === "forEach")
        return true;
      return false;
    });
    if (!target)
      return ctx.reportError("Unable to find .forEach() to convert");
    const member = target.callee;
    const iterator = member.object;
    const fn = target.arguments[0];
    if (fn.type !== "ArrowFunctionExpression" && fn.type !== "FunctionExpression")
      return ctx.reportError("Unable to find function to convert");
    if (fn.params.length !== 1) {
      return ctx.reportError(
        "Unable to convert forEach",
        {
          node: fn.params[0],
          message: "Index argument in forEach is not yet supported for conversion"
        }
      );
    }
    const returnNodes = [];
    ctx.traverse(fn.body, (path, { SKIP }) => {
      if (FOR_TRAVERSE_IGNORE.includes(path.node.type))
        return SKIP;
      if (path.node.type === "ReturnStatement")
        returnNodes.push(path.node);
    });
    let textBody = ctx.source.getText(fn.body);
    returnNodes.sort((a, b) => b.loc.start.line - a.loc.start.line).forEach((c) => {
      textBody = textBody.slice(0, c.range[0] - fn.body.range[0]) + "continue" + textBody.slice(c.range[1] - fn.body.range[0]);
    });
    const local = fn.params[0];
    const str = `for (const ${ctx.source.getText(local)} of ${ctx.source.getText(iterator)}) ${textBody}`;
    ctx.removeComment();
    ctx.report({
      node: target,
      message: "Convert to for-of loop",
      fix(fixer) {
        return fixer.replaceText(target, str);
      }
    });
  }
};

const commands = {
  keepSorted,
  toArrow,
  toForEach,
  toForOf,
  toFunction
};

const SKIP = Symbol("skip");
const STOP = Symbol("stop");
function traverse(context, node, visitor) {
  const allVisitorKeys = context.sourceCode.visitorKeys;
  const queue = [];
  queue.push({
    node,
    parent: null,
    parentKey: null,
    parentPath: null
  });
  while (queue.length) {
    const currentPath = queue.shift();
    const result = visitor(currentPath, { SKIP, STOP });
    if (result === STOP)
      return false;
    if (result === SKIP)
      continue;
    const visitorKeys = allVisitorKeys[currentPath.node.type];
    if (!visitorKeys)
      continue;
    for (const visitorKey of visitorKeys) {
      const child = currentPath.node[visitorKey];
      if (!child) {
        continue;
      } else if (Array.isArray(child)) {
        for (const item of child) {
          queue.push({
            node: item,
            parent: currentPath.node,
            parentKey: visitorKey,
            parentPath: currentPath
          });
        }
      } else {
        queue.push({
          node: child,
          parent: currentPath.node,
          parentKey: visitorKey,
          parentPath: currentPath
        });
      }
    }
  }
  return true;
}

function createRuleWithCommands(commands2) {
  return createEslintRule({
    name: "command",
    meta: {
      type: "problem",
      docs: {
        description: "Enforce top-level functions to be declared with function keyword",
        recommended: "stylistic"
      },
      fixable: "code",
      schema: [],
      messages: {
        "command-error": "[{{command}}] error: {{message}}",
        "command-error-cause": "[{{command}}] error cause: {{message}}",
        "command-fix": "[{{command}}] fix: {{message}}",
        "command-removal": "[{{command}}] remove comment after use"
      }
    },
    defaultOptions: [],
    create: (context) => {
      const sc = context.sourceCode;
      const comments = sc.getAllComments();
      for (const comment of comments) {
        if (comment.type !== "Line")
          continue;
        const commandRaw = comment.value.trim();
        for (const command of commands2) {
          if (command.match.test(commandRaw)) {
            command.action(new CommandContextImpl(context, comment, command));
            continue;
          }
        }
      }
      return {};
    }
  });
}
class CommandContextImpl {
  constructor(context, comment, command) {
    this.context = context;
    this.comment = comment;
    this.command = command;
  }
  get source() {
    return this.context.sourceCode;
  }
  reportError(message, cause) {
    this.context.report({
      loc: this.comment.loc,
      messageId: "command-error",
      data: {
        command: this.command.name,
        message
      }
    });
    if (cause) {
      const { message: message2, ...pos } = cause;
      this.context.report(
        {
          ...pos,
          messageId: "command-error-cause",
          data: {
            command: this.command.name,
            message: message2
          }
        }
      );
    }
  }
  report({ message, ...report }) {
    this.context.report({
      ...report,
      messageId: "command-fix",
      data: {
        command: this.command.name,
        message,
        ...report.data
      }
    });
  }
  traverse(node, cb) {
    return traverse(this.context, node, cb);
  }
  removeComment() {
    this.context.report({
      loc: this.comment.loc,
      messageId: "command-removal",
      data: {
        command: this.command.name
      },
      fix: (fixer) => {
        const lastToken = this.context.sourceCode.getTokenBefore(
          this.comment,
          { includeComments: true }
        )?.range[1];
        let lineStart = this.context.sourceCode.getIndexFromLoc({
          line: this.comment.loc.start.line,
          column: 0
        }) - 1;
        if (lastToken != null)
          lineStart = Math.max(lastToken, lineStart);
        return fixer.removeRange([
          lineStart,
          this.comment.range[1]
        ]);
      }
    });
  }
  findNodeBelow(...keys) {
    const tokenBelow = this.context.sourceCode.getTokenAfter(this.comment);
    if (!tokenBelow)
      return;
    const nodeBelow = this.context.sourceCode.getNodeByRangeIndex(tokenBelow.range[1]);
    if (!nodeBelow)
      return;
    let result;
    let target = nodeBelow;
    while (target.parent && target.parent.loc.start.line === nodeBelow.loc.start.line)
      target = target.parent;
    const filter = typeof keys[0] === "function" ? keys[0] : (node) => keys.includes(node.type);
    this.traverse(target, (path) => {
      if (path.node.loc.start.line !== nodeBelow.loc.start.line)
        return STOP;
      if (filter(path.node)) {
        result = path.node;
        return STOP;
      }
    });
    return result;
  }
}
const command = createRuleWithCommands(Object.values(commands));

function createPluginWithCommands(options = {}) {
  const {
    name = "command"
  } = options;
  return {
    meta: {
      name,
      version
    },
    rules: {
      command
    }
  };
}

const defaultPlugin = createPluginWithCommands();

export { createPluginWithCommands as c, defaultPlugin as d };
